#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
"""RAMSES RF - a RAMSES-II protocol decoder & analyser.

A pseudo-mocked serial port used for testing.
"""

__dev_mode__ = False

MOCKED_PORT = "/dev/ttyMOCK"

GWY_ID = "18:181818"
